/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.hibiscus.mahoe.assistant;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.rmi.RemoteException;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.frankmuenster.hibiscus.mahoe.aktien.assistant.AktienAccountMigrator;
import de.frankmuenster.hibiscus.mahoe.bsh.assistant.BshAccountMigrator;
import de.frankmuenster.hibiscus.mahoe.santander.assistant.SantanderAccountMigrator;
import de.frankmuenster.hibiscus.mahoe.targobank.assistant.TargoBankAccountMigrator;
import de.frankmuenster.hibiscus.mahoe.utils.I18N;
import de.willuhn.datasource.rmi.DBIterator;
import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.messaging.Message;
import de.willuhn.jameica.messaging.MessageConsumer;
import de.willuhn.jameica.messaging.SystemMessage;
import de.willuhn.jameica.plugin.Plugin;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Started den Account-Migrator erst nach vollstndigem Systemstart
 *
 * @author Frank Mnster
 */
public class MahoeUpdateHandler implements MessageConsumer {

    /**
     * Liste mit allen bekannten AccountMigratoren/Updatern
     */
    private final transient Set<Class<? extends AccountMigrator>> accountMigrators = new HashSet<>();

    /**
     * Erstellt die Migrator-Klasse und registriert sich selbst in der MessagingFactory
     */
    public MahoeUpdateHandler() {

        this.accountMigrators.add(AktienAccountMigrator.class);
        this.accountMigrators.add(BshAccountMigrator.class);
        this.accountMigrators.add(SantanderAccountMigrator.class);
        this.accountMigrators.add(TargoBankAccountMigrator.class);
    }

    /**
     * @see de.willuhn.jameica.messaging.MessageConsumer#autoRegister()
     */
    @Override
    public boolean autoRegister() {

        return true;
    }

    /**
     * @see de.willuhn.jameica.messaging.MessageConsumer#getExpectedMessageTypes()
     */
    @SuppressWarnings("rawtypes")
    @Override
    public Class[] getExpectedMessageTypes() {

        return new Class[] { SystemMessage.class };
    }

    /**
     * @throws ApplicationException
     * @see de.willuhn.jameica.messaging.MessageConsumer#handleMessage(de.willuhn.jameica.messaging.Message)
     */
    @Override
    public void handleMessage(final Message message) throws ApplicationException {

        final SystemMessage sysMsg = (SystemMessage) message;
        if (sysMsg.getStatusCode() == SystemMessage.SYSTEM_STARTED) {

            Logger.debug("System gestartet. Nun kann der Migrator gestartet werden."); //$NON-NLS-1$
            this.checkOldPlugins();
            this.updateAccounts();
            this.unregister();
        }

    }

    /**
     * @throws ApplicationException
     */
    private void checkOldPlugins() throws ApplicationException {

        final List<Plugin> plugins = Application.getPluginLoader().getInstalledPlugins();

        for (final Class<? extends AccountMigrator> clazz : this.accountMigrators) {

            try {
                final Constructor<? extends AccountMigrator> cons = clazz.getConstructor();
                final AccountMigrator accountMigrator = cons.newInstance();
                for (final Plugin plugin : plugins) {
                    accountMigrator.migrate(plugin);
                }
            } catch (NoSuchMethodException | SecurityException | InstantiationException | IllegalAccessException
                    | IllegalArgumentException | InvocationTargetException e) {
                Logger.error(I18N.tr("MahoeUpdateHandler.checkforOldScriptingModulesFailed", e.getMessage()), e); //$NON-NLS-1$
            }
        }

    }

    private void unregister() {

        // Unregister
        Application.getMessagingFactory().unRegisterMessageConsumer(this);
    }

    private void updateAccounts() throws ApplicationException {

        for (final Class<? extends AccountMigrator> clazz : this.accountMigrators) {

            try {
                final Constructor<? extends AccountMigrator> cons = clazz.getConstructor();
                final AccountMigrator accountMigrator = cons.newInstance();

                final HBCIDBService hcbiService = Settings.getDBService();
                final DBIterator<Konto> accounts = hcbiService.createList(Konto.class);
                while (accounts.hasNext()) {
                    final Konto account = accounts.next();
                    accountMigrator.update(account);
                }
            } catch (NoSuchMethodException | SecurityException | InstantiationException | IllegalAccessException
                    | IllegalArgumentException | InvocationTargetException | RemoteException e) {
                Logger.error(I18N.tr("MahoeUpdateHandler.checkforOldScriptingModulesFailed", e.getMessage()), e); //$NON-NLS-1$
            }
        }

    }

}
