
package de.frankmuenster.hibiscus.mahoe.messages;

import java.rmi.RemoteException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import com.gargoylesoftware.htmlunit.WebClient;

import de.frankmuenster.hibiscus.mahoe.utils.I18N;
import de.frankmuenster.hibiscus.mahoe.web.WebSiteException;
import de.willuhn.datasource.rmi.DBIterator;
import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.hbci.rmi.HBCIDBService;
import de.willuhn.jameica.hbci.rmi.Nachricht;
import de.willuhn.jameica.messaging.StatusBarMessage;
import de.willuhn.jameica.system.Application;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Abstrakte-Klasse die Bank-Nachrichten verarbeitet
 *
 * @author Frank Mnster
 */
public abstract class AbstractBankMessages implements InterfaceBankMessages {

    /**
     * Logger information in case an method is not yet implemented
     */
    public static final String NOT_IMPLEMENTED = "Noch nicht implementiert."; //$NON-NLS-1$

    /**
     * Der WebClient
     */
    protected WebClient webclient;

    /**
     * Liste mit Bank-Nachrichten
     */
    protected List<MahoeMessage> messages;

    /**
     * berprft, ob die Nachricht schon existiert.
     *
     * @param blz
     * @param date
     * @param subject
     * @return true, wenn die Nachricht bereits vorhanden ist, false sonst.
     */
    public static boolean existsMessage(final String blz, final Date date, final String subject) {

	final SimpleDateFormat sdf = new SimpleDateFormat("dd.MM.yyyy HH:mm", Locale.GERMAN); //$NON-NLS-1$
	final String date2 = sdf.format(date);
	final String keyToTest = date2 + subject;
	Logger.debug("Key der neuen Nachricht: " + keyToTest); //$NON-NLS-1$
	boolean exist = false;
	try {
	    final DBIterator<Nachricht> messages = AbstractBankMessages.getBankMessages(blz);
	    while (messages.hasNext()) {
		final Nachricht msg = messages.next();
		final String keyFormDb = msg.getMeta("doubleKey", StringUtils.EMPTY); //$NON-NLS-1$
		Logger.debug("Key der gespeicherten Nachricht: " + keyToTest); //$NON-NLS-1$
		if (keyFormDb.equals(keyToTest)) {
		    exist = true;
		    break;
		}
	    }
	} catch (final RemoteException e) {
	    Logger.error(I18N.tr("AbstractBankMessages.compareFailed"), e); //$NON-NLS-1$
	}
	return exist;
    }

    /**
     * Liefert die Nachrichtenliste fr eine Bank
     *
     * @param blz
     *            die Bankleitzahl
     * @return die Liste mit den Nachrichten
     * @throws RemoteException
     */
    private static DBIterator<Nachricht> getBankMessages(final String blz) throws RemoteException {

	final HBCIDBService service = Settings.getDBService();
	final DBIterator<Nachricht> list = service.createList(Nachricht.class);
	list.addFilter("blz=?", blz); //$NON-NLS-1$
	return list;
    }

    /**
     * Konstruktor mit Angabe des initialisierten WebClient
     *
     * @param webClient
     */
    public AbstractBankMessages(final WebClient webClient) {
	this.webclient = webClient;
	this.messages = new ArrayList<>();
    }

    /**
     * @see de.frankmuenster.hibiscus.mahoe.messages.InterfaceBankMessages#
     *      readMessages (java.util.Date)
     * @throws WebSiteException
     * @throws ApplicationException
     */
    @Override
    public abstract void readMessages(final Date startDate) throws WebSiteException, ApplicationException;

    /**
     * @see de.frankmuenster.hibiscus.mahoe.messages.InterfaceBankMessages#
     *      saveMessage ()
     */
    @Override
    public int saveMessage() throws ApplicationException {

	int savedMsg = 0;

	for (final MahoeMessage mahoeMsg : this.messages) {
	    try {

		mahoeMsg.saveAsNachricht();
		this.markAsRead(mahoeMsg);

		final String text = I18N.tr("AbstractBankMessages.newBankMessage"); //$NON-NLS-1$
		Application.getMessagingFactory()
			.sendMessage(new StatusBarMessage(text, StatusBarMessage.TYPE_SUCCESS)); // NOPMD
		savedMsg++;
	    } catch (final RemoteException e) {
		Logger.error(I18N.tr("AbstractBankMessages.storeFailed"), e); //$NON-NLS-1$
		// Im Fehlerfall zeigen wir einfach den Dialog an
		try {
		    Application.getCallback().notifyUser(mahoeMsg.getMessage());
		} catch (final Exception e1) { // NOPMD
		    throw new ApplicationException(I18N.tr("AbstractBankMessages.viewFailed"), e1); // NOPMD //$NON-NLS-1$
		}
	    }
	}

	return savedMsg;
    }
}
