/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.hibiscus.mahoe.santander.web;

import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.apache.commons.lang.StringUtils;

import com.gargoylesoftware.htmlunit.BrowserVersion;
import com.gargoylesoftware.htmlunit.ElementNotFoundException;
import com.gargoylesoftware.htmlunit.FailingHttpStatusCodeException;
import com.gargoylesoftware.htmlunit.NicelyResynchronizingAjaxController;
import com.gargoylesoftware.htmlunit.WebClient;
import com.gargoylesoftware.htmlunit.WebClientOptions;
import com.gargoylesoftware.htmlunit.html.DomElement;
import com.gargoylesoftware.htmlunit.html.HtmlAnchor;
import com.gargoylesoftware.htmlunit.html.HtmlForm;
import com.gargoylesoftware.htmlunit.html.HtmlInput;
import com.gargoylesoftware.htmlunit.html.HtmlPage;
import com.gargoylesoftware.htmlunit.html.HtmlPasswordInput;
import com.gargoylesoftware.htmlunit.html.HtmlSubmitInput;
import com.gargoylesoftware.htmlunit.html.HtmlTable;
import com.gargoylesoftware.htmlunit.html.HtmlTableRow;

import de.frankmuenster.hibiscus.mahoe.utils.I18N;
import de.frankmuenster.hibiscus.mahoe.utils.Utils;
import de.frankmuenster.hibiscus.mahoe.utils.pdfsaver.PdfSaverException;
import de.frankmuenster.hibiscus.mahoe.web.AbstractWebSite;
import de.frankmuenster.hibiscus.mahoe.web.WebSiteException;
import de.frankmuenster.hibiscus.mahoe.web.WebSiteExceptionTypes;
import de.willuhn.jameica.hbci.rmi.AuslandsUeberweisung;
import de.willuhn.jameica.hbci.rmi.Konto;
import de.willuhn.jameica.hbci.rmi.KontoType;
import de.willuhn.jameica.hbci.rmi.Umsatz;
import de.willuhn.logging.Logger;
import de.willuhn.util.ApplicationException;

/**
 * Santander Bank Website.
 *
 * @author Frank Mnster
 * @param <I>
 *            SantanderWebAccountInformation
 * @param <M>
 *            SantanderBankMessages
 */
@SuppressWarnings({ "PMD.ExcessiveImports" })
public abstract class AbstractSantanderWebSite<I extends SantanderWebAccountInformation, M extends SantanderBankMessages>
        extends AbstractWebSite<I, M> {

    /**
     * Anchor details anchor
     */
    protected static final int ANCHOR_CELL_DETAILS = 3;

    /**
     * Anchor details anchor
     */
    protected static final int ANCHOR_CELL_TRANSACTIONS = 0;

    /** Unsere BceeWebSite instance. */
    private static AbstractSantanderWebSite<SantanderWebAccountInformation, SantanderBankMessages> webSite;

    /** Der Benutzername. */
    private final String userName;

    /**
     * Liefert die Instance der WebSeite.
     *
     * @param konto
     *            the konto
     * @return die WebSeite
     * @throws WebSiteException
     *             the web site exception
     * @throws ApplicationException
     *             the application exception
     */
    public static AbstractSantanderWebSite<SantanderWebAccountInformation, SantanderBankMessages> getInstance(
            final Konto konto) throws WebSiteException, ApplicationException {

        String userName = null;
        KontoType type = null;
        try {
            userName = konto.getKundennummer();
            type = KontoType.find(konto.getAccountType());
        } catch (final RemoteException e) {
            final String msg = I18N.tr("AbstractSantanderWebSite.loginMissing"); //$NON-NLS-1$
            Logger.error(msg, e);
            throw new ApplicationException(msg, e);
        }
        if (AbstractSantanderWebSite.webSite == null) {
            AbstractSantanderWebSite.createNewInstance(konto, type, userName);
        } else {
            // Ist der Benutzername und der Kontotyp derselbe
            if (AbstractSantanderWebSite.webSite.userName.equals(userName)
                    && AbstractSantanderWebSite.webSite.isSupporting(type)) {
                // Aktualisiere Startdatum der Umsatzabfrage
                final Date lastTransacDate = Utils.getLastUmsatzDate(konto,
                        AbstractSantanderWebSite.webSite.getStartdate());
                AbstractSantanderWebSite.webSite.accountInfo.setLastTransactionDate(lastTransacDate);
            } else {
                AbstractSantanderWebSite.webSite.logoff();
                AbstractSantanderWebSite.createNewInstance(konto, type, userName);
            }
        }
        return AbstractSantanderWebSite.webSite;
    }

    /**
     * Creates the new instance.
     *
     * @param konto
     *            the konto
     * @param type
     *            the type
     * @param userName
     *            the user name
     * @throws ApplicationException
     *             the application exception
     * @throws WebSiteException
     *             the web site exception
     */
    private static void createNewInstance(final Konto konto, final KontoType type, final String userName)
            throws ApplicationException, WebSiteException {

        if (KontoType.GIRO.equals(type)) {
            AbstractSantanderWebSite.webSite = new SantanderWebSiteGiro(konto, userName);
        } else if (KontoType.KREDITKARTE.equals(type)) {
            AbstractSantanderWebSite.webSite = new SantanderWebSiteKreditkarte(konto, userName);
        } else {
            throw new ApplicationException(I18N.tr("AbstractSantanderWebSite.missingAccountType")); //$NON-NLS-1$
        }

    }

    /**
     * Erstellt ein neues Webseiten-Objekt zur Kontaktaufnahme mit der BCEE.
     *
     * @param konto
     *            the konto
     * @param userName
     *            the user name
     * @throws WebSiteException
     *             the web site exception
     */
    protected AbstractSantanderWebSite(final Konto konto, final String userName) throws WebSiteException {

        super(konto);
        this.userName = userName;
        this.accountInfo.setBankType(SantanderBankEnum.find(this.accountInfo.getKonto()));
        this.webClient = this.getWebClient();
    }

    /**
     * Creates the sepa transfer.
     *
     * @param aUeberweisung
     *            the a ueberweisung
     * @throws WebSiteException
     *             the web site exception
     * @throws ApplicationException
     *             the application exception
     * @see de.frankmuenster.hibiscus.mahoe.web.WebSite#createSepaTransfer(de.willuhn.jameica.hbci.rmi.AuslandsUeberweisung)
     */
    @Override
    public void createSepaTransfer(final AuslandsUeberweisung aUeberweisung)
            throws WebSiteException, ApplicationException {

        Logger.info(I18N.tr("AbstractSantanderWebSite.transferNotSupported")); //$NON-NLS-1$
    }

    /**
     * Gets the account information.
     *
     * @return the account information
     * @throws WebSiteException
     *             the web site exception
     * @throws ApplicationException
     *             the application exception
     * @see de.frankmuenster.hibiscus.mahoe.web.WebSite#getAccountInformation()
     */
    @Override
    public abstract I getAccountInformation() throws WebSiteException, ApplicationException;

    /**
     * Gets the booked transactions.
     *
     * @return the booked transactions
     * @throws WebSiteException
     *             the web site exception
     * @throws ApplicationException
     *             the application exception
     * @see de.frankmuenster.hibiscus.mahoe.web.WebSite#getBookedTransactions()
     */
    @Override
    public abstract List<Umsatz> getBookedTransactions() throws WebSiteException, ApplicationException;

    /**
     * Gets the pre booked transactions.
     *
     * @return the pre booked transactions
     * @throws WebSiteException
     *             the web site exception
     * @throws ApplicationException
     *             the application exception
     * @see de.frankmuenster.hibiscus.mahoe.web.WebSite#getPreBookedTransactions()
     */
    @Override
    public abstract List<Umsatz> getPreBookedTransactions() throws WebSiteException, ApplicationException;

    /**
     * Liefert einen instanzierten HtmlUnit WebClient.
     *
     * @return der WebClient
     */
    @Override
    public final WebClient getWebClient() {

        if (this.webClient == null) {
            this.webClient = new WebClient(BrowserVersion.FIREFOX_52);
            final WebClientOptions options = this.webClient.getOptions();
            options.setUseInsecureSSL(false);
            options.setJavaScriptEnabled(true);
            options.setThrowExceptionOnFailingStatusCode(false);
            options.setCssEnabled(true);
            options.setThrowExceptionOnScriptError(false);
            options.setRedirectEnabled(true);
            this.webClient.setAjaxController(new NicelyResynchronizingAjaxController());
            this.webClient.getCookieManager().setCookiesEnabled(true);
            Utils.setProxyCfg(this.webClient, SantanderBankEnum.CONSUMER.getLoginUrl());
        }
        return this.webClient;
    }

    /**
     * Checks if is still logged in.
     *
     * @return true, if is still logged in
     * @throws WebSiteException
     *             the web site exception
     * @see de.frankmuenster.hibiscus.mahoe.web.WebSite#isStillLoggedIn()
     */
    @Override
    public boolean isStillLoggedIn() throws WebSiteException {

        // Brauchen wir noch nicht, Wird halt neu angemeldet.
        return false;
    }

    /**
     * Login.
     *
     * @param password
     *            the password
     * @param loginParams
     *            muss Login_URL enthalten
     * @throws WebSiteException
     *             the web site exception
     * @see de.frankmuenster.hibiscus.mahoe.web.WebSite#login(char[], java.lang.String[])
     */
    @Override
    public void login(final char[] password, final String... loginParams) throws WebSiteException {

        if (loginParams.length != 0) {
            throw new WebSiteException(I18N.tr("AbstractSantanderWebSite.loginParameterNumberWrong"), //$NON-NLS-1$
                    WebSiteExceptionTypes.LOGIN_ERROR);
        }
        // Login URL
        final String loginURL = SantanderBankEnum.CONSUMER.getLoginUrl();
        try {
            final HtmlPage pageLogin = this.webClient.getPage(loginURL);
            final HtmlForm formLogin = pageLogin.getFormByName("loginDataForm"); //$NON-NLS-1$
            Logger.debug(formLogin.asText());
            final HtmlInput inputUser = formLogin.getInputByName("loginDataForm.personenNummer"); //$NON-NLS-1$
            final HtmlPasswordInput inputPasswd = formLogin.getInputByName("loginDataForm.pin"); //$NON-NLS-1$
            final HtmlInput submitContinue = formLogin.getInputByValue("Login"); //$NON-NLS-1$
            // Setze Benutzernamen
            Logger.debug("Setzte Kundenummer auf: ****" + this.userName.substring(this.userName.length() - 4)); //$NON-NLS-1$
            inputUser.setValueAttribute(this.userName);
            // Setze Passwort
            inputPasswd.setValueAttribute(new String(password));
            // Invoke Click
            final HtmlPage pageContent = submitContinue.click();
            // And again test for error
            if (pageContent.asText().contains("Letzter Login") //$NON-NLS-1$
                    || pageContent.asText().contains("Letzte erfolgreiche Anmeldung")) { //$NON-NLS-1$
                Logger.info(I18N.tr("AbstractSantanderWebSite.loginSuccessful")); //$NON-NLS-1$
            } else if (pageContent.asText()
                    .contains("aufgrund einer Wartung steht der Internetservice kurzzeitig nicht zur")) { //$NON-NLS-1$
                Logger.info(I18N.tr("AbstractSantanderWebSite.webServiceUnavailable")); //$NON-NLS-1$
                throw new WebSiteException(I18N.tr("AbstractSantanderWebSite.webServiceUnavailable"), //$NON-NLS-1$
                        WebSiteExceptionTypes.SERVICE_UNAVAILABLE);
            } else {
                Logger.error(I18N.tr("AbstractSantanderWebSite.loginFailed")); //$NON-NLS-1$
                throw new WebSiteException(I18N.tr("AbstractSantanderWebSite.loginFailed"), //$NON-NLS-1$
                        WebSiteExceptionTypes.LOGIN_ERROR);
            }
        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(I18N.tr("AbstractSantanderWebSite.loginFailedLog", e.getMessage()), e); //$NON-NLS-1$
            throw new WebSiteException(I18N.tr("AbstractSantanderWebSite.openLoginPageFailed"), //$NON-NLS-1$
                    WebSiteExceptionTypes.LOGIN_ERROR, e);
        }
    }

    /**
     * Logoff.
     *
     * @throws WebSiteException
     *             the web site exception
     * @see de.frankmuenster.hibiscus.mahoe.web.WebSite#logoff()
     */
    @Override
    public void logoff() throws WebSiteException {

        try {
            // Switch of SSL check as this can lead to a hand-shake error here
            this.webClient.getOptions().setUseInsecureSSL(true);
            final HtmlPage page = this.webClient.getPage(SantanderBankEnum.CONSUMER.getOverviewUrl());
            final HtmlForm logoffForm = page.getFormByName("LOGOFF"); //$NON-NLS-1$
            final HtmlSubmitInput submit = logoffForm.getInputByValue("Abmelden"); //$NON-NLS-1$
            final HtmlPage logoffPage = submit.click();
            // Switch on SecureSSL again
            this.webClient.getOptions().setUseInsecureSSL(false);
            Logger.trace(logoffPage.asText());
            if (!logoffPage.asText().contains("Sie wurden erfolgreich vom Online Banking abgemeldet.")) { //$NON-NLS-1$
                Logger.error(I18N.tr("AbstractSantanderWebSite.loggoutFailed")); //$NON-NLS-1$
            }
        } catch (ElementNotFoundException | FailingHttpStatusCodeException | IOException e) {
            // Nur Loggen
            Logger.error(I18N.tr("AbstractSantanderWebSite.loggoutFailed"), e); //$NON-NLS-1$
        }
    }

    /**
     * Save PDF.
     *
     * @param accountInfo
     *            the account info
     * @return the int
     * @throws PdfSaverException
     *             the pdf saver exception
     * @see de.frankmuenster.hibiscus.mahoe.web.WebSite#savePDF(de.frankmuenster.hibiscus.mahoe.web.WebAccountInformation)
     */
    @Override
    public abstract int savePDF(final SantanderWebAccountInformation accountInfo) throws PdfSaverException;

    /**
     * Returns the anchor to the account details page for the given account type (Giro/credit card) and number
     *
     * @param accountNumber
     *            the account number
     * @param tableIdentifier
     *            the table identifier for the type of accounts
     * @param anchorCell
     *            the cell number that contains the anchor
     * @return the anchor to the details page or <code>null</code>
     * @throws WebSiteException
     *             if the overview page could not be loaded.
     */
    @SuppressWarnings("PMD.OnlyOneReturn")
    protected HtmlAnchor getAccountDetailsAnchor(final String accountNumber, final String tableIdentifier,
            final int anchorCell) throws WebSiteException {

        try {
            // open overview page
            final HtmlPage overviewOPage = this.webClient.getPage(SantanderBankEnum.CONSUMER.getOverviewUrl());

            // get all tables in page
            final List<DomElement> tables = overviewOPage.getElementsByTagName(AbstractWebSite.TABLE_TAG); // $NON-NLS-1$
            for (final DomElement domElement : tables) {
                // Loop through tables
                final HtmlTable htmlTable = (HtmlTable) domElement;
                Logger.debug(htmlTable.getFirstChild().getNextSibling().getTextContent());
                if (tableIdentifier.equals(htmlTable.getFirstChild().getNextSibling().getTextContent())) {
                    // we are on the correct table, now search account anchor
                    final List<HtmlTableRow> rows = htmlTable.getBodies().get(0).getRows();
                    for (final HtmlTableRow row : rows) {
                        String cellZero = row.getCell(0).asText().toLowerCase(Locale.GERMAN);
                        cellZero = cellZero.replaceAll("\\s", StringUtils.EMPTY); //$NON-NLS-1$
                        if (cellZero.contains(accountNumber)) {
                            // OK, we found the anchor, now break and return
                            if (row.getCell(anchorCell).getFirstElementChild() instanceof HtmlAnchor) {
                                return (HtmlAnchor) row.getCell(anchorCell).getFirstElementChild();
                            } else {
                                return (HtmlAnchor) row.getCell(anchorCell).getFirstElementChild()
                                        .getFirstElementChild();
                            }
                        }
                    }
                }
            }

        } catch (FailingHttpStatusCodeException | IOException e) {
            Logger.error(I18N.tr("SantanderWebSiteGiro.readAccountOverviewFailedLog", //$NON-NLS-1$
                    e.getMessage()), e);
            throw new WebSiteException(I18N.tr("SantanderWebSiteGiro.readAccountOverviewFailed"), // NOPMD //$NON-NLS-1$
                    WebSiteExceptionTypes.READ_ACCOUNT_DETAILS_FAILED, e);
        }

        return null;

    }

    /**
     * Gets the startdate.
     *
     * @return the startdate
     * @see de.frankmuenster.hibiscus.mahoe.web.AbstractWebSite#getStartdate()
     */
    @Override
    protected abstract Calendar getStartdate();

    /**
     * &Uuml;perf&uuml;pt, ob die WebSeiteninstance den Kontotypen unterst&uuml;tzt.
     *
     * @param type
     *            der Kontoyp
     * @return <i>true</i>, wenn der Kontotyp unterst&uuml;tzt wird, <i>false</i> sonst.
     */
    protected abstract boolean isSupporting(KontoType type);
}
