/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/
package de.frankmuenster.hibiscus.mahoe.targobank.pdfextractor;

import java.io.IOException;
import java.io.InputStream;
import java.text.NumberFormat;
import java.util.List;
import java.util.Locale;

import org.apache.pdfbox.text.PDFTextStripper;

import de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractBankPDFPage;
import de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor;
import de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractTransactionCreator;
import de.frankmuenster.hibiscus.mahoe.pdfextractor.PdfBuchung;
import de.frankmuenster.hibiscus.mahoe.pdfextractor.PdfEintrag;
import de.willuhn.logging.Logger;

/**
 * Class for extracting entry information from a TargoBank PDF bank statement.
 *
 * @author Frank Mnster
 */
public class TargoBankPdfExtract extends AbstractPdfExtractor {

    /**
     * Der der die letzte Seite mit Umstzen angibt
     */
    private static final String LAST_PAGE_TEXT = "Endsaldo:"; //$NON-NLS-1$

    /**
     * Zeilenhhe der Buchungszeile im MahoeMain PDF.
     */
    private static final double LINE_HEIGH = 9.35;

    /**
     * Dezimalformatierer
     */
    private static final NumberFormat N_FORMAT = NumberFormat.getInstance(Locale.GERMAN);

    /**
     * Erstellt einen PDF-Extraktor der TargoBank
     *
     * @see AbstractPdfExtractor#AbstractPdfExtractor(InputStream, String)
     */
    public TargoBankPdfExtract(final InputStream inputStream, final String kartenNummer) {

        super(inputStream, kartenNummer);
    }

    /**
     * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#removeAnnotationsBookingDates(java.util.List)
     */
    @Override
    protected void removeAnnotationsBookingDates(final List<PdfEintrag> buchungenDatum) {

        if (!buchungenDatum.isEmpty()) {
            final double xOfFirstDatum = buchungenDatum.get(0).getPositionX();
            Logger.trace("First date: " + xOfFirstDatum); //$NON-NLS-1$
            for (int i = buchungenDatum.size() - 1; i > 0; i--) {
                Logger.trace("Check date x position: " + buchungenDatum.get(i).getPositionX()); //$NON-NLS-1$
                if (buchungenDatum.get(i).getPositionX() != xOfFirstDatum) {
                    Logger.trace("Remove date as it is an annotation in the bottom"); //$NON-NLS-1$
                    buchungenDatum.remove(i);
                }
            }
        }
    }

    /**
     * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#removeAnnotations(java.util.List)
     */
    @Override
    protected void removeAnnotations(final List<PdfEintrag> pdfWortList) {

        final double yOfFirstDatum = this.buchungsDaten.get(this.buchungsDaten.size() - 1).getPositionY() - 10;
        Logger.debug("First: " + yOfFirstDatum); //$NON-NLS-1$
        for (int i = pdfWortList.size() - 1; i >= 0; i--) {
            Logger.debug("Check: " + pdfWortList.get(i).getPositionY()); //$NON-NLS-1$
            if (pdfWortList.get(i).getPositionY() < yOfFirstDatum) {
                Logger.debug("Remove"); //$NON-NLS-1$
                pdfWortList.remove(i);
            }
        }
    }

    /**
     * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#addYear
     */
    @Override
    protected void addYear(final List<PdfEintrag> entries) {

        // Jahr hinzu tun
        final int rMonat = Integer.parseInt(this.rechungsDatum.substring(3, 5));
        final int rJahr = Integer.parseInt(this.rechungsDatum.substring(6, 10));
        for (final PdfEintrag entry : entries) {
            // Bei Partnerkarten kommt es zu fehlerhaften Eintrgen, die hier zu
            // einer Exception fhren
            try {
                final int bMonat = Integer.parseInt(entry.getWort().substring(3, 5));
                if (bMonat > rMonat) {
                    entry.setWort(entry.getWort() + (rJahr - 1));
                } else {
                    entry.setWort(entry.getWort() + rJahr);
                }
            } catch (final NumberFormatException e) { // NOPMD
                // ignorieren wir
            }
        }
    }

    /**
     * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getTransactionCreator()
     */
    @Override
    protected AbstractTransactionCreator getTransactionCreator() {

        return new TargoBankTransactionCreator(this.buchungsDaten, TargoBankPdfExtract.LINE_HEIGH);
    }

    /**
     * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getNumberFormat()
     */
    @Override
    protected NumberFormat getNumberFormat() {

        return TargoBankPdfExtract.N_FORMAT;
    }

    /**
     * @throws IOException
     * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getBankPdfPageFirst()
     */
    @Override
    protected AbstractBankPDFPage getBankPdfPageFirst() throws IOException {

        return new TargoBankPdfPageFirst();
    }

    /**
     * @throws IOException
     * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getBankPdfPageEven()
     */
    @Override
    protected AbstractBankPDFPage getBankPdfPageEven() throws IOException {

        return new TargoBankPdfPageEven();
    }

    /**
     * @throws IOException
     * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getBankPdfPageOdd()
     */
    @Override
    protected AbstractBankPDFPage getBankPdfPageOdd() throws IOException {

        return new TargoBankPdfPageOdd();
    }

    /**
     * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#reOrder(java.util.List)
     */
    @Override
    protected void reOrder(final List<PdfBuchung> buchungen) {

        // Leider mssen wir richtig sortieren, da die Umstze nach Karten
        // sortiert sind
        buchungen.sort((final PdfBuchung buchung1,
                final PdfBuchung buchung2) -> buchung1.getBuchungsDatum().compareTo(buchung2.getBuchungsDatum()) * -1);
    }

    /**
     * @throws IOException
     * @see de.frankmuenster.hibiscus.mahoe.pdfextractor.AbstractPdfExtractor#getLastPage()
     */
    @Override
    protected int getLastPage() throws IOException {

        int lastPage = this.document.getPages().getCount();
        final PDFTextStripper textStripper = new PDFTextStripper();
        for (int pageNumber = 1; pageNumber <= this.document.getPages().getCount(); pageNumber++) {
            textStripper.setStartPage(pageNumber);
            textStripper.setEndPage(pageNumber);
            // Fetch page text and insert into "pages" string
            final String pageText = textStripper.getText(this.document);
            if (pageText.contains(TargoBankPdfExtract.LAST_PAGE_TEXT)) {
                lastPage = pageNumber;
                break;
            }
        }
        return lastPage;
    }
}
