/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.hibiscus.mahoe.utils;

import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.security.AlgorithmParameters;
import java.security.GeneralSecurityException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;

import javax.crypto.Cipher;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.SecretKeySpec;
import javax.xml.bind.DatatypeConverter;

import de.willuhn.logging.Logger;

/**
 * Ein Crypto-Klasse zum Salten und Hashen von Passwrtern. Kann nur innerhalb einer Session genutzt werden, nicht zum
 * permanentem Speichern, da sich der Schlssel bei jedem Start ndert.
 *
 * @author Frank Mnster
 * @version 1.0
 */
public final class Crypto {

    /**
     * Iterationanzahl
     */
    private static int iterations = 10000;

    /**
     * Verschlsselungsstrke (nur 128, da hhere Schlssel nicht 'out-of-box' von Java untersttz werden.
     */
    private static int keySize = 128;

    /**
     * Initial Variant
     */
    private static IvParameterSpec ivBytes;

    /**
     * Und natrlich der geheime Schlssel
     */
    // private static SecretKey secretKey;

    /**
     * Und auch den Secret Key
     */
    private static SecretKeySpec secretSpec;

    /**
     * Und die Cipher-Instanze
     */
    private static Cipher cipher;

    private Crypto() {

        throw new AssertionError("Instantiating utility class..."); //$NON-NLS-1$
    }

    /**
     * Verschlsselt einen Text als char array im Default encoding.
     *
     * @param plaintext
     *            char[] der zu verschlsselnde Text
     * @return String der verschlsselte Text
     * @throws GeneralSecurityException
     *             falls irgend ein Fehler auftritt
     */
    public static String encrypt(final char[] plaintext) // NOPMD
            throws GeneralSecurityException {

        if (plaintext == null) {
            throw new GeneralSecurityException(I18N.tr("Crypto.argumentCannotBeNull")); //$NON-NLS-1$
        }

        byte[] encryptedText;

        synchronized (Crypto.class) {
            if (Crypto.secretSpec == null) {
                Logger.debug("Erstelle secretKey"); //$NON-NLS-1$
                final byte[] saltBytes = Crypto.getSalt().getBytes();
                final SecretKeyFactory skf = SecretKeyFactory.getInstance("PBKDF2WithHmacSHA1"); //$NON-NLS-1$
                final PBEKeySpec spec = new PBEKeySpec(plaintext, saltBytes, Crypto.iterations, Crypto.keySize);
                final SecretKey secretKey = skf.generateSecret(spec);
                Crypto.secretSpec = new SecretKeySpec(secretKey.getEncoded(), "AES"); //$NON-NLS-1$
            }

            if (Crypto.cipher == null) {
                Logger.debug("Erstelle ivBytes"); //$NON-NLS-1$
                Crypto.cipher = Cipher.getInstance("AES/CBC/PKCS5Padding"); //$NON-NLS-1$
                Crypto.cipher.init(Cipher.ENCRYPT_MODE, Crypto.secretSpec);
                final AlgorithmParameters params = Crypto.cipher.getParameters();
                Crypto.ivBytes = params.getParameterSpec(IvParameterSpec.class);
            }
            Crypto.cipher.init(Cipher.ENCRYPT_MODE, Crypto.secretSpec, Crypto.ivBytes);
        }

        final Charset uft8 = Charset.forName(Charset.defaultCharset().name());
        final ByteBuffer byteBuffer = uft8.encode(CharBuffer.wrap(plaintext));
        encryptedText = Crypto.cipher.doFinal(byteBuffer.array());

        return DatatypeConverter.printBase64Binary(encryptedText);
    }

    /**
     * Entschlsselt einen verschlsselten Text.
     *
     * @param encryptedText
     *            der verschlsselte Text
     * @return char[] der entschlsselte Text.
     * @throws GeneralSecurityException
     *             , falls irgendwas schief geht.
     */
    public static char[] decrypt(final String encryptedText) throws GeneralSecurityException {

        if (encryptedText == null) {
            throw new GeneralSecurityException(I18N.tr("Crypto.argumentCannotBeNull")); //$NON-NLS-1$
        }
        CharBuffer charBuffer;

        final byte[] encryptTextBytes = DatatypeConverter.parseBase64Binary(encryptedText);
        Crypto.cipher.init(Cipher.DECRYPT_MODE, Crypto.secretSpec, Crypto.ivBytes);

        final byte[] decryptedText = Crypto.cipher.doFinal(encryptTextBytes);

        final Charset uft8 = Charset.forName(Charset.defaultCharset().name());
        charBuffer = uft8.decode(ByteBuffer.wrap(decryptedText));

        return charBuffer.array();

    }

    private static String getSalt() throws NoSuchAlgorithmException {

        final SecureRandom sRandom = SecureRandom.getInstance("SHA1PRNG"); //$NON-NLS-1$
        final byte[] salt = new byte[20];
        sRandom.nextBytes(salt);
        return new String(salt);
    }

}
