/*******************************************************************************
 * Copyright 2015 - 2018 fmCreation
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald diese von der
 * Europischen Kommission genehmigt wurden - Folgeversionen der EUPL
 * ("Lizenz"); Sie drfen dieses Werk ausschlielich gem dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 *   https://progs.frankmuenster.de/eupl_v1-2_de
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften gefordert oder in
 * schriftlicher Form vereinbart, wird die unter der Lizenz verbreitete Software
 * "so wie sie ist", OHNE JEGLICHE GEWHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrcklich oder stillschweigend - verbreitet. Die sprachspezifischen
 * Genehmigungen und Beschrnkungen unter der Lizenz sind dem Lizenztext zu
 * entnehmen.
 ******************************************************************************/

package de.frankmuenster.hibiscus.mahoe.utils;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import de.willuhn.jameica.hbci.Settings;
import de.willuhn.jameica.security.Wallet;
import de.willuhn.jameica.system.Application;
import de.willuhn.jameica.system.OperationCanceledException;
import de.willuhn.logging.Logger;

/**
 * Ein Password-Cache inclusive Hibiscus Wallet Benutzung.
 *
 * @author Frank Mnster
 * @version 1.0
 */
public final class PasswdCache {

    /**
     *
     */
    private static final String DELETE_FROM_WALLET_FAILED = "PasswdCache.deletePwFromWalletFailed"; //$NON-NLS-1$

    /**
     * Der Walletkey Prefix
     */
    private static final String KEY_PREFIX = "mahoe_"; //$NON-NLS-1$

    /**
     * Die HashMap fr die Passwrter
     */
    private static Map<String, String> passwdMap;

    /**
     * Passwortabfrage abgebrochen Text
     */
    private static final String ASK_PW_CANCELLED = "PasswdCache.askPwCancelled"; //$NON-NLS-1$

    /**
     * Passwortabfrage Dialog ffnen Fehlertext
     */
    private static final String ASK_PW_FAILED = "PasswdCache.openRequestDialogFailed"; //$NON-NLS-1$
    /**
     * Legt die Hashmap statisch an.
     */
    static {
	PasswdCache.passwdMap = new HashMap<>();
    }

    /**
     * Liefert das Passwort, die PIN. Sucht im Cache bzw. Hibiscus-Wallet, falls
     * das konfiguriert ist. Fragt den Benutzer, falls das Passwort noch nicht
     * gespeichert ist, bzw. speichern nicht erlaubt ist.
     *
     * @param bank
     *            die Bankbezeichnung
     * @param login
     *            - der Login zum Passwort
     * @return das Passwort
     * @throws OperationCanceledException
     */
    public static char[] getPassword(final String bank, final String login) throws OperationCanceledException {

	return PasswdCache.getPassword(bank, login, false);
    }

    /**
     * Liefert das Passwort, die PIN. Sucht im Cache bzw. Hibiscus-Wallet, falls
     * das konfiguriert ist. Fragt den Benutzer, falls das Passwort noch nicht
     * gespeichert ist, bzw. speichern nicht erlaubt ist.
     *
     * @param bank
     *            die Bankbezeichnung
     * @param login
     *            der Login zum Passwort
     * @param alwaysAsk
     *            immer nach dem Passwort fragen, egal was in den Einstellungen
     *            gesetzt ist.
     * @return das Passwort
     * @throws OperationCanceledException
     */
    public static char[] getPassword(final String bank, final String login, final boolean alwaysAsk)
	    throws OperationCanceledException {

	Boolean cachePin = false;
	Boolean storePin = false;
	if (!alwaysAsk) {
	    // Should PIN be cached during session?
	    cachePin = Settings.getCachePin();
	    Logger.info(I18N.tr("PasswdCache.savePinInSession", cachePin.toString())); //$NON-NLS-1$
	    // Should PIN be stored permanently?
	    storePin = Settings.getStorePin();
	    Logger.info(I18N.tr("PasswdCache.savePinPermanent", storePin.toString())); //$NON-NLS-1$
	}
	// Try to get session password
	char[] passwd = null;
	if (cachePin && storePin) {
	    passwd = PasswdCache.cacheAndStore(bank, login);
	} else if (cachePin) {
	    passwd = PasswdCache.cacheOnly(bank, login);
	} else if (storePin) {
	    passwd = PasswdCache.storeOnly(bank, login);
	} else {
	    passwd = PasswdCache.askAlways(bank, login);
	}
	return passwd == null ? null : Arrays.copyOf(passwd, passwd.length);
    }

    /**
     * Entfernt ein Passwort aus dem Cache
     *
     * @param bank
     *            der Bankname
     * @param login
     *            das Login/die Kundenkennung
     */
    public static void revoke(final String bank, final String login) {

	final String passwordKey = PasswdCache.KEY_PREFIX + bank + "_" + login; //$NON-NLS-1$
	PasswdCache.revoke(passwordKey);
    }

    private static char[] askAlways(final String bank, final String login) {

	char[] passwd = new char[0];
	// delete cache and store
	PasswdCache.revoke(login);
	final String passwordKey = PasswdCache.KEY_PREFIX + bank + "_" + login; //$NON-NLS-1$
	// Reset previous stored password
	try {
	    Wallet hibiscusWallet = null;
	    hibiscusWallet = Settings.getWallet();
	    hibiscusWallet.set(passwordKey, null);
	} catch (final Exception e) { // NOPMD
	    Logger.warn(I18N.tr(PasswdCache.DELETE_FROM_WALLET_FAILED));
	}
	try {
	    passwd = PasswdCache.askuser(bank, login);
	} catch (final OperationCanceledException eo) {
	    PasswdCache.revoke(passwordKey);
	    Logger.warn(I18N.tr(PasswdCache.ASK_PW_CANCELLED));
	    throw eo;
	} catch (final Exception e) { // NOPMD
	    Logger.error(I18N.tr(PasswdCache.ASK_PW_FAILED, e.getMessage()), e);
	}
	return passwd;
    }

    private static char[] askuser(final String bank, final String login) throws Exception { // NOPMD

	final String pWord = Application.getCallback().askPassword(I18N.tr("PasswdCache.requestPw", login, bank)); //$NON-NLS-1$
	return pWord.toCharArray();
    }

    private static char[] cacheAndStore(final String bank, final String login) {

	char[] passwd = null;
	final String passwordKey = PasswdCache.KEY_PREFIX + bank + "_" + login; //$NON-NLS-1$
	try {
	    // Lese Passwort aus Cache
	    if (PasswdCache.passwdMap.containsKey(passwordKey)) {
		passwd = Crypto.decrypt(PasswdCache.passwdMap.get(passwordKey));
	    }
	    if (passwd == null) {

		// Lese Passwort aus Wallet
		final Wallet hibiscusWallet = Settings.getWallet();
		passwd = (char[]) hibiscusWallet.get(passwordKey);

		if (PasswdCache.isBlank(passwd)) {
		    // Benutzer fragen und dann in Cache speichern
		    passwd = PasswdCache.askuser(bank, login);
		    hibiscusWallet.set(passwordKey, passwd);
		}
		PasswdCache.passwdMap.put(passwordKey, Crypto.encrypt(passwd));
	    } else {
		// Nur zur Sicherheit in Wallet speichern
		final Wallet hibiscusWallet = Settings.getWallet();
		hibiscusWallet.set(passwordKey, passwd);
	    }
	} catch (final OperationCanceledException eo) {
	    PasswdCache.revoke(passwordKey);
	    Logger.warn(I18N.tr(PasswdCache.ASK_PW_CANCELLED));
	    throw eo;
	} catch (final Exception e) { // NOPMD
	    Logger.error(I18N.tr(PasswdCache.ASK_PW_FAILED, e.getMessage()), e);
	}
	return passwd;
    }

    private static char[] cacheOnly(final String bank, final String login) {

	char[] passwd = null;
	final String passwordKey = PasswdCache.KEY_PREFIX + bank + "_" + login; //$NON-NLS-1$
	// Reset previous stored password
	try {
	    Wallet hibiscusWallet = null;
	    hibiscusWallet = Settings.getWallet();
	    hibiscusWallet.set(passwordKey, null);
	} catch (final Exception e) { // NOPMD
	    Logger.warn(I18N.tr(PasswdCache.DELETE_FROM_WALLET_FAILED));
	}
	try {
	    // Lese Passwort aus Cache
	    if (PasswdCache.passwdMap.containsKey(passwordKey)) {
		passwd = Crypto.decrypt(PasswdCache.passwdMap.get(passwordKey));
	    }
	    if (passwd == null) {
		// Benutzer fragen und dann in Cache speichern
		passwd = PasswdCache.askuser(bank, login);
		PasswdCache.passwdMap.put(passwordKey, Crypto.encrypt(passwd));
	    }
	} catch (final OperationCanceledException eo) {
	    PasswdCache.revoke(passwordKey);
	    Logger.warn(I18N.tr(PasswdCache.ASK_PW_CANCELLED));
	    throw eo;
	} catch (final Exception e) { // NOPMD
	    Logger.error(I18N.tr(PasswdCache.ASK_PW_FAILED, e.getMessage()), e);
	}
	return passwd;
    }

    /**
     * Checks if a char sequence is not empty (""), not null and not whitespace
     * only.
     *
     * @param charSequence
     *            the char sequence to check, maybe null
     * @return <code>true</code> if the String is not empty and not null and not
     *         whitespace
     */
    private static boolean isBlank(final char... charSequence) {

	boolean isBlank = true;
	if (charSequence != null) {
	    for (final char c : charSequence) {
		if (!Character.isWhitespace(c)) {
		    isBlank = false;
		    break;
		}
	    }
	}
	return isBlank;
    }

    /**
     * Entfernt ein Passwort aus dem Cache.
     *
     * @param passwordKey
     */
    private static void revoke(final String passwordKey) {

	try {
	    if (PasswdCache.passwdMap.containsKey(passwordKey)) {
		PasswdCache.passwdMap.remove(passwordKey);
	    }
	    final Wallet hibiscusWallet = Settings.getWallet();
	    hibiscusWallet.set(passwordKey, null);
	} catch (final Exception e) { // NOPMD
	    Logger.warn(I18N.tr(PasswdCache.DELETE_FROM_WALLET_FAILED));
	}
    }

    private static char[] storeOnly(final String bank, final String login) {

	char[] passwd = new char[0];
	final String passwordKey = PasswdCache.KEY_PREFIX + bank + "_" + login; //$NON-NLS-1$
	// Remove from cache if exists
	if (PasswdCache.passwdMap.containsKey(passwordKey)) {
	    PasswdCache.passwdMap.remove(passwordKey);
	}
	try {
	    // Lese Passwort aus Wallet lesen
	    final Wallet hibiscusWallet = Settings.getWallet();
	    passwd = (char[]) hibiscusWallet.get(passwordKey);
	    if (passwd == null) {
		// Benutzer fragen und dann in Cache speichern
		passwd = PasswdCache.askuser(bank, login);
		hibiscusWallet.set(passwordKey, passwd);
	    }
	} catch (final OperationCanceledException eo) {
	    PasswdCache.revoke(passwordKey);
	    Logger.warn(I18N.tr(PasswdCache.ASK_PW_CANCELLED));
	    throw eo;
	} catch (final Exception e) { // NOPMD
	    Logger.error(I18N.tr(PasswdCache.ASK_PW_FAILED, e.getMessage()), e);
	}
	return passwd;
    }

    /**
     * Privater Konstruktor
     */
    private PasswdCache() {

	throw new AssertionError("Instantiating utility class..."); //$NON-NLS-1$
    }
}
